/******************************************************************************

  ble.c 

  driver for BLE protocol configuration and operation

******************************************************************************/

#include <stdint.h>
#include <string.h>
#include "nordic_common.h"
#include "nrf.h"
#include "ble_hci.h"
#include "ble_advdata.h"
#include "ble_advertising.h"
#include "ble_conn_params.h"
#include "nrf_sdh.h"
#include "nrf_sdh_soc.h"
#include "nrf_sdh_ble.h"
#include "nrf_ble_gatt.h"
#include "app_timer.h"
#include "ble_tj.h"
#include "app_uart.h"
#include "app_util_platform.h"
#include "bsp_btn_ble.h"
#include "ble_pump.h"  

#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"

#define MAN_DATA_LEN    24

#define AUTH_DELAY  (APP_TIMER_CLOCK_FREQ * 60) // one minute delay given the fact that we are using RTC timer

#define APP_BLE_CONN_CFG_TAG            1                                           /**< A tag identifying the SoftDevice BLE configuration. */

#define APP_FEATURE_NOT_SUPPORTED       BLE_GATT_STATUS_ATTERR_APP_BEGIN + 2        /**< Reply when unsupported features are requested. */

#define DEVICE_NAME                     "TJ-Pump"                               /**< Name of device. Will be included in the advertising data. */
#define TJ_SERVICE_UUID_TYPE           BLE_UUID_TYPE_VENDOR_BEGIN                  /**< UUID type for the Nordic UART Service (vendor specific). */

#define APP_BLE_OBSERVER_PRIO           1                                           /**< Application's BLE observer priority. You shouldn't need to modify this value. */

#define APP_ADV_INTERVAL                100                                          /**< The advertising interval (in units of 0.625 ms. This value corresponds to 40 ms). */
#define APP_ADV_TIMEOUT_IN_SECONDS      180                                         /**< The advertising timeout (in units of seconds). */

#define MIN_CONN_INTERVAL               MSEC_TO_UNITS(20, UNIT_1_25_MS)             /**< Minimum acceptable connection interval (20 ms), Connection interval uses 1.25 ms units. */
#define MAX_CONN_INTERVAL               MSEC_TO_UNITS(75, UNIT_1_25_MS)             /**< Maximum acceptable connection interval (75 ms), Connection interval uses 1.25 ms units. */
#define SLAVE_LATENCY                   0                                           /**< Slave latency. */
#define CONN_SUP_TIMEOUT                MSEC_TO_UNITS(4000, UNIT_10_MS)             /**< Connection supervisory timeout (4 seconds), Supervision Timeout uses 10 ms units. */
#define FIRST_CONN_PARAMS_UPDATE_DELAY  APP_TIMER_TICKS(5000)                       /**< Time from initiating event (connect or start of notification) to first time sd_ble_gap_conn_param_update is called (5 seconds). */
#define NEXT_CONN_PARAMS_UPDATE_DELAY   APP_TIMER_TICKS(30000)                      /**< Time between each call to sd_ble_gap_conn_param_update after the first call (30 seconds). */
#define MAX_CONN_PARAMS_UPDATE_COUNT    3                                           /**< Number of attempts before giving up the connection parameter negotiation. */

#define TJ_COMPANY_ID          (0xCAFE)

#define DEAD_BEEF                       0xDEADBEEF                                  /**< Value used as error code on stack dump, can be used to identify stack location on stack unwind. */

/**@brief  ADV state capture
*
* @details This structure is stores internal adv state
*/
typedef struct
{
  bool adv_active;
  ble_adv_mode_t ble_adv_state;   /** <Current ADV state > */
  uint32_t ble_adv_map;       /** < advertising channel map currently used> */
  int16_t  ble_adv_tx_power;     /** < Current TX power used for advertising > */
  ble_advdata_manuf_data_t  adv_man_data;       /** < Current adv manufacture data used for ADV > */
  uint8_t  ble_adv_man_data[MAN_DATA_LEN];   /** < Current MAN data used for ADV > */
  uint32_t     adv_interval;
} ble_adv_state_t;



typedef struct
{
  uint16_t   conn_handle;
  uint16_t   ble_tj_max_data_len;
  bool     connection_authenticated;
  bool     response_registered;
  bool     alert_registered;
  ble_gap_addr_t peer_addr;
  
  app_tj_data_handler_t p_dataCB;
  app_tj_state_handler_t p_StateCB;
  app_tj_tx_completed_handler_t p_TxCompletedCB;
  app_tj_rssi_crossed_t			p_rssiChangedCB;
  app_tj_auth_handler_t			p_authHandlerCB;
} ble_connection_state_t;




BLE_TJ_DEF(m_tj);                                                                 /**< TJ service instance. */
NRF_BLE_GATT_DEF(m_gatt);                                                           /**< GATT module instance. */
BLE_ADVERTISING_DEF(m_advertising);                                                 /**< Advertising module instance. */
APP_TIMER_DEF(m_auth_timer_id);                                                  /**< Security Request timer. */
#if TEST_BLE_ACK 
APP_TIMER_DEF(m_test_timer_id);                                                  /**< Security Request timer. */
static void rssiCB(int8_t rssiv);
#endif


static ble_adv_state_t    ble_tj_adv_state;
static ble_connection_state_t ble_tj_conn_state;

#ifdef __DATA_TEST__
static uint8_t received_data[512];
static uint16_t received_data_length;
#endif

/**@brief Function for handling an event from the Connection Parameters Module.
*
* @details This function will be called for all events in the Connection Parameters Module
*          which are passed to the application.
*
* @note All this function does is to disconnect. This could have been done by simply setting
*       the disconnect_on_fail config parameter, but instead we use the event handler
*       mechanism to demonstrate its use.
*
* @param[in] p_evt  Event received from the Connection Parameters Module.
*/
static void on_conn_params_evt(ble_conn_params_evt_t * p_evt)
{
  uint32_t err_code;
  
  if (p_evt->evt_type == BLE_CONN_PARAMS_EVT_FAILED)
  {
    err_code = sd_ble_gap_disconnect(ble_tj_conn_state.conn_handle, BLE_HCI_CONN_INTERVAL_UNACCEPTABLE);
    APP_ERROR_CHECK(err_code);
  }
}


/**@brief Function for handling errors from the Connection Parameters module.
*
* @param[in] nrf_error  Error code containing information about what went wrong.
*/
static void conn_params_error_handler(uint32_t nrf_error)
{
  APP_ERROR_HANDLER(nrf_error);
}

/**@brief Function for initializing the Connection Parameters module.
*/
static void conn_params_init(void)
{
  uint32_t               err_code;
  ble_conn_params_init_t cp_init;
  
  memset(&cp_init, 0, sizeof(cp_init));
  
  cp_init.p_conn_params                  = NULL;
  cp_init.first_conn_params_update_delay = FIRST_CONN_PARAMS_UPDATE_DELAY;
  cp_init.next_conn_params_update_delay  = NEXT_CONN_PARAMS_UPDATE_DELAY;
  cp_init.max_conn_params_update_count   = MAX_CONN_PARAMS_UPDATE_COUNT;
  cp_init.start_on_notify_cccd_handle    = BLE_GATT_HANDLE_INVALID;
  cp_init.disconnect_on_fail             = false;
  cp_init.evt_handler                    = on_conn_params_evt;
  cp_init.error_handler                  = conn_params_error_handler;
  
  err_code = ble_conn_params_init(&cp_init);
  APP_ERROR_CHECK(err_code);
}


/**@brief Function for handling events from the GATT library. */
void gatt_evt_handler(nrf_ble_gatt_t * p_gatt, nrf_ble_gatt_evt_t const * p_evt)
{
  if ((ble_tj_conn_state.conn_handle == p_evt->conn_handle) && (p_evt->evt_id == NRF_BLE_GATT_EVT_ATT_MTU_UPDATED))
  {
    ble_tj_conn_state.ble_tj_max_data_len = p_evt->params.att_mtu_effective - OPCODE_LENGTH - HANDLE_LENGTH;
    NRF_LOG_INFO("Data len is set to 0x%X(%d)",  ble_tj_conn_state.ble_tj_max_data_len,  ble_tj_conn_state.ble_tj_max_data_len);
  }
  NRF_LOG_DEBUG("ATT MTU exchange completed. central 0x%x peripheral 0x%x",
                p_gatt->att_mtu_desired_central,
                p_gatt->att_mtu_desired_periph);
}


/**@brief Function for initializing the GATT library. */
static void gatt_init(void)
{
  ret_code_t err_code;
  
  err_code = nrf_ble_gatt_init(&m_gatt, gatt_evt_handler);
  APP_ERROR_CHECK(err_code);
  
  err_code = nrf_ble_gatt_att_mtu_periph_set(&m_gatt, NRF_SDH_BLE_GATT_MAX_MTU_SIZE);
  APP_ERROR_CHECK(err_code);
}


/**@brief Function for handling BLE events.
*
* @param[in]   p_ble_evt   Bluetooth stack event.
* @param[in]   p_context   Unused.
*/
static void ble_evt_handler(ble_evt_t const * p_ble_evt, void * p_context)
{
  uint32_t err_code;
  
  switch (p_ble_evt->header.evt_id)
  {
  case BLE_GAP_EVT_CONNECTED:
    NRF_LOG_INFO("Connected");
#if HW_REV==HW_REV_NORDIC_EVAL_10040_S132
    err_code = bsp_indication_set(BSP_INDICATE_CONNECTED);
    APP_ERROR_CHECK(err_code);
#endif            
    ble_tj_adv_state.adv_active = false;
    ble_tj_conn_state.conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
    ble_tj_conn_state.peer_addr = p_ble_evt->evt.gap_evt.params.connected.peer_addr;
	
    // ARM Authentication timeout
    app_timer_start(m_auth_timer_id, AUTH_DELAY, NULL);
    break;
    
  case BLE_GAP_EVT_DISCONNECTED:
    NRF_LOG_INFO("Disconnected");
#if HW_REV==HW_REV_NORDIC_EVAL_10040_S132
    err_code = bsp_indication_set(BSP_INDICATE_IDLE);
    APP_ERROR_CHECK(err_code);
#endif            
    
	// If registered to the RSSI notifications - unregister from them
	if (ble_tj_conn_state.p_rssiChangedCB)
	{
		sd_ble_gap_rssi_stop(ble_tj_conn_state.conn_handle);
		ble_tj_conn_state.p_rssiChangedCB = 0;
	}

	
	ble_tj_conn_state.conn_handle = BLE_CONN_HANDLE_INVALID;
	ble_tj_conn_state.connection_authenticated = false;
	ble_tj_conn_state.response_registered = false;
	ble_tj_conn_state.alert_registered = false;
	memset(&ble_tj_conn_state.peer_addr, 0, sizeof(ble_tj_conn_state.peer_addr));

	// Restart ADV
	BLE_do_adv_start(ble_tj_adv_state.ble_adv_map, ble_tj_adv_state.ble_adv_tx_power);
    if (0)
    {
      ble_tj_conn_state.p_StateCB(BLE_STATE_DISCONNECTED);
    }
    break;
    
  case BLE_GAP_EVT_PHY_UPDATE_REQUEST:
    {
      NRF_LOG_DEBUG("PHY update request.");
      ble_gap_phys_t const phys =
      {
        .rx_phys = BLE_GAP_PHY_AUTO,
        .tx_phys = BLE_GAP_PHY_AUTO,
      };
      err_code = sd_ble_gap_phy_update(p_ble_evt->evt.gap_evt.conn_handle, &phys);
      APP_ERROR_CHECK(err_code);
    } break;
    
  case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
    // Pairing not supported
    err_code = sd_ble_gap_sec_params_reply(ble_tj_conn_state.conn_handle, BLE_GAP_SEC_STATUS_PAIRING_NOT_SUPP, NULL, NULL);
    APP_ERROR_CHECK(err_code);
    break;
    
  case BLE_GAP_EVT_DATA_LENGTH_UPDATE_REQUEST:
    {
      ble_gap_data_length_params_t dl_params;
      
      // Clearing the struct will effectivly set members to @ref BLE_GAP_DATA_LENGTH_AUTO
      memset(&dl_params, 0, sizeof(ble_gap_data_length_params_t));
      err_code = sd_ble_gap_data_length_update(p_ble_evt->evt.gap_evt.conn_handle, &dl_params, NULL);
      APP_ERROR_CHECK(err_code);
    } break;
	
  case BLE_GAP_EVT_RSSI_CHANGED:
    {
		if (ble_tj_conn_state.p_rssiChangedCB)
		{
			ble_tj_conn_state.p_rssiChangedCB(p_ble_evt->evt.gap_evt.params.rssi_changed.rssi);
		}
  	}
	break;
  
  case BLE_GATTS_EVT_SYS_ATTR_MISSING:
    // No system attributes have been stored.
    err_code = sd_ble_gatts_sys_attr_set(ble_tj_conn_state.conn_handle, NULL, 0, 0);
    APP_ERROR_CHECK(err_code);
    break;
    
  case BLE_GATTC_EVT_TIMEOUT:
    // Disconnect on GATT Client timeout event.
    err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gattc_evt.conn_handle,
                                     BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
    APP_ERROR_CHECK(err_code);
    break;
    
  case BLE_GATTS_EVT_TIMEOUT:
    // Disconnect on GATT Server timeout event.
    err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gatts_evt.conn_handle,
                                     BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
    APP_ERROR_CHECK(err_code);
    break;
    
  case BLE_EVT_USER_MEM_REQUEST:
    err_code = sd_ble_user_mem_reply(p_ble_evt->evt.gattc_evt.conn_handle, NULL);
    APP_ERROR_CHECK(err_code);
    break;
    
  case BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST:
    {
      ble_gatts_evt_rw_authorize_request_t  req;
      ble_gatts_rw_authorize_reply_params_t auth_reply;
      
      req = p_ble_evt->evt.gatts_evt.params.authorize_request;
      
      if (req.type != BLE_GATTS_AUTHORIZE_TYPE_INVALID)
      {
        if ((req.request.write.op == BLE_GATTS_OP_PREP_WRITE_REQ)     ||
            (req.request.write.op == BLE_GATTS_OP_EXEC_WRITE_REQ_NOW) ||
              (req.request.write.op == BLE_GATTS_OP_EXEC_WRITE_REQ_CANCEL))
        {
          if (req.type == BLE_GATTS_AUTHORIZE_TYPE_WRITE)
          {
            auth_reply.type = BLE_GATTS_AUTHORIZE_TYPE_WRITE;
          }
          else
          {
            auth_reply.type = BLE_GATTS_AUTHORIZE_TYPE_READ;
          }
          auth_reply.params.write.gatt_status = APP_FEATURE_NOT_SUPPORTED;
          err_code = sd_ble_gatts_rw_authorize_reply(p_ble_evt->evt.gatts_evt.conn_handle,
                                                     &auth_reply);
          APP_ERROR_CHECK(err_code);
        }
      }
    } 
    break; // BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST
    
  default:
    // No implementation needed.
    break;
  }
}


/**@brief Function for the SoftDevice initialization.
*
* @details This function initializes the SoftDevice and the BLE event interrupt.
*/
static void ble_stack_init(void)
{
  ret_code_t err_code;
  
  err_code = nrf_sdh_enable_request();
  APP_ERROR_CHECK(err_code);
  
  // Configure the BLE stack using the default settings.
  // Fetch the start address of the application RAM.
  uint32_t ram_start = 0;
  err_code = nrf_sdh_ble_default_cfg_set(APP_BLE_CONN_CFG_TAG, &ram_start);
  APP_ERROR_CHECK(err_code);
  
  // Enable BLE stack.
  err_code = nrf_sdh_ble_enable(&ram_start);
  APP_ERROR_CHECK(err_code);
  
  // Register a handler for BLE events.
  NRF_SDH_BLE_OBSERVER(m_ble_observer, APP_BLE_OBSERVER_PRIO, ble_evt_handler, NULL);
}

/**@brief Function for handling advertising events.
*
* @details This function will be called for advertising events which are passed to the application.
*
* @param[in] ble_adv_evt  Advertising event.
*/
static void on_adv_evt(ble_adv_evt_t ble_adv_evt)
{
#if HW_REV==HW_REV_NORDIC_EVAL_10040_S132
  uint32_t err_code;
#endif
  
  switch (ble_adv_evt)
  {
  case BLE_ADV_EVT_SLOW:
  case BLE_ADV_EVT_FAST:
#if HW_REV==HW_REV_NORDIC_EVAL_10040_S132
    err_code = bsp_indication_set(BSP_INDICATE_ADVERTISING);
    APP_ERROR_CHECK(err_code);
#endif            
    ble_tj_adv_state.adv_active = true;
    if (ble_tj_conn_state.p_StateCB)
    {
      ble_tj_conn_state.p_StateCB(BLE_STATE_ADV);
    }
    break;
  case BLE_ADV_EVT_IDLE:
    BLE_do_adv_start(ble_tj_adv_state.ble_adv_map, ble_tj_adv_state.ble_adv_tx_power);
    break;
  default:
    break;
  }
}

/**@brief Function for initializing the Advertising functionality.
*/
static void advertising_init(uint32_t interval, bool first)
{
  uint32_t               err_code;
  ble_advertising_init_t init;
  
  memset(&init, 0, sizeof(init));
  
  if (first)
  {
    memset(&ble_tj_adv_state, 0, sizeof(ble_tj_adv_state));
  }
  
  ble_tj_adv_state.adv_interval   = interval;
  init.advdata.name_type          = BLE_ADVDATA_NO_NAME;
  init.advdata.include_appearance = false;
  init.advdata.flags              = BLE_GAP_ADV_FLAGS_LE_ONLY_LIMITED_DISC_MODE;
  
  init.srdata.uuids_complete.uuid_cnt = 0;
  init.srdata.uuids_complete.p_uuids  = 0;
  
  
  if (interval <= APP_ADV_INTERVAL)
  {
    init.config.ble_adv_fast_enabled  = true;
    init.config.ble_adv_slow_enabled  = false;
    init.config.ble_adv_fast_interval = interval;
    init.config.ble_adv_fast_timeout  = APP_ADV_TIMEOUT_IN_SECONDS;
    
    ble_tj_adv_state.ble_adv_state = BLE_ADV_MODE_FAST;
  }
  else
  {
    init.config.ble_adv_fast_enabled  = false;
    init.config.ble_adv_slow_enabled  = true;
    init.config.ble_adv_slow_interval = interval;
    init.config.ble_adv_slow_timeout  = APP_ADV_TIMEOUT_IN_SECONDS;
    
    ble_tj_adv_state.ble_adv_state = BLE_ADV_MODE_SLOW;
  }
  
  init.evt_handler = on_adv_evt;
  
  
  // ADV MAN_ID and manufacture specific data
  init.advdata.p_manuf_specific_data = &ble_tj_adv_state.adv_man_data;
  
  init.advdata.p_manuf_specific_data->company_identifier = TJ_COMPANY_ID;
  init.advdata.p_manuf_specific_data->data.size = sizeof(ble_tj_adv_state.ble_adv_man_data);
  init.advdata.p_manuf_specific_data->data.p_data = ble_tj_adv_state.ble_adv_man_data;
  
  err_code = ble_advertising_init(&m_advertising, &init);
  APP_ERROR_CHECK(err_code);
  
  ble_advertising_conn_cfg_tag_set(&m_advertising, APP_BLE_CONN_CFG_TAG);
}

/**@brief Function for the GAP initialization.
*
* @details This function will set up all the necessary GAP (Generic Access Profile) parameters of
*          the device. It also sets the permissions and appearance.
*/
static void gap_params_init(uint16_t  connection_interval_min,
                            uint16_t  connection_interval_max,
                            uint16_t  slave_latency,
                            uint16_t  supervision_timeout_,
                            int16_t  tx_power)
{
  uint32_t                err_code;
  ble_gap_conn_params_t   gap_conn_params;
  ble_gap_conn_sec_mode_t sec_mode;
  
  BLE_GAP_CONN_SEC_MODE_SET_OPEN(&sec_mode);
  
  err_code = sd_ble_gap_device_name_set(&sec_mode,
                                        (const uint8_t *) DEVICE_NAME,
                                        strlen(DEVICE_NAME));
  APP_ERROR_CHECK(err_code);
  
  memset(&gap_conn_params, 0, sizeof(gap_conn_params));
  
  gap_conn_params.min_conn_interval = MIN_CONN_INTERVAL;
  gap_conn_params.max_conn_interval = MAX_CONN_INTERVAL;
  gap_conn_params.slave_latency     = SLAVE_LATENCY;
  gap_conn_params.conn_sup_timeout  = CONN_SUP_TIMEOUT;
  
  err_code = sd_ble_gap_ppcp_set(&gap_conn_params);
  APP_ERROR_CHECK(err_code);
}

/**@brief Function for handling the data from the Nordic TJ Service.
*
* @details This function will process the data received from the TJ service which is either 
*      authentication data or command
*
* @param[in] p_tj    TJ Service structure.
* @param[in] p_data   Data to be send to UART module.
* @param[in] length   Length of the data.
*/
/**@snippet [Handling the data received over BLE] */
static void tj_data_handler(ble_tj_evt_t * p_evt)
{ 
  if (p_evt->type == BLE_TJ_EVT_RX_DATA)
  {
    if (ble_tj_conn_state.connection_authenticated == true)
    {
      NRF_LOG_DEBUG("Received data from Controller.");
      NRF_LOG_HEXDUMP_DEBUG(p_evt->params.rx_data.p_data, p_evt->params.rx_data.length);
      
      if (ble_tj_conn_state.p_dataCB)
      {
        ble_tj_conn_state.p_dataCB(p_evt->params.rx_data.p_data, p_evt->params.rx_data.length);
      }
      
#ifdef __DATA_TEST__
      received_data_length = p_evt->params.rx_data.length;
      memcpy(received_data, p_evt->params.rx_data.p_data, received_data_length);
      
      
      app_timer_start(m_test_timer_id, (APP_TIMER_CLOCK_FREQ * 1), NULL);
#endif
    }
    
  }
  else if (p_evt->type == BLE_TJ_EVT_RX_AUTH)
  {
    if (ble_tj_conn_state.connection_authenticated == false)
    {
    	bool auth = false;
		
    	if (ble_tj_conn_state.p_authHandlerCB) 
		{
			auth = ble_tj_conn_state.p_authHandlerCB(ble_tj_conn_state.peer_addr.addr, (uint8_t *)p_evt->params.rx_data.p_data, p_evt->params.rx_data.length);
    	}

      if (auth)
      {
		// Authentication succeeded - cancel authentication timer
		app_timer_stop(m_auth_timer_id);
		ble_tj_conn_state.connection_authenticated = true;
		// TBD: store MAC address of the connection into the NVMEM as paired device


		if (ble_tj_conn_state.p_StateCB && (ble_tj_conn_state.alert_registered) 
			&&  (ble_tj_conn_state.response_registered))
		{
		  ble_tj_conn_state.p_StateCB(BLE_STATE_CONNECTED);

#if TEST_BLE_ACK
		BLE_callback_rssi_threshold_crossed(0, rssiCB);

		app_timer_start(m_test_timer_id, (APP_TIMER_CLOCK_FREQ * 2), NULL);
#endif
        }
      }
	  else
	  {
	  	// Authentication failed - disconnect connection upon authentication timeout, which will fire soon 
	  }
    }
    else
    {
      NRF_LOG_DEBUG("Authentication attempts while already authenticated");
    }
  }
  else if (p_evt->type == BLE_TJ_EVT_TX_RDY)
  {
    // Notify that TX completed
    if (ble_tj_conn_state.p_TxCompletedCB)
    {
      ble_tj_conn_state.p_TxCompletedCB();
    }
    
  }
  else if (p_evt->type == BLE_TJ_EVT_RSP_STARTED)
  {
  	ble_tj_conn_state.response_registered = true;
  	if  (ble_tj_conn_state.connection_authenticated == true)
  	{
  		if (ble_tj_conn_state.alert_registered)
  		{
  			if (ble_tj_conn_state.p_StateCB)
        	{
	        	ble_tj_conn_state.p_StateCB(BLE_STATE_CONNECTED);

#if TEST_BLE_ACK
				BLE_callback_rssi_threshold_crossed(0, rssiCB);

				app_timer_start(m_test_timer_id, (APP_TIMER_CLOCK_FREQ * 2), NULL);
#endif
	        }
  		}
  	}
  }
  else if (p_evt->type == BLE_TJ_EVT_EVENT_STARTED)
	{
		ble_tj_conn_state.alert_registered = true;
	  	if  (ble_tj_conn_state.connection_authenticated == true)
	  	{
	  		if (ble_tj_conn_state.response_registered)
	  		{
	  			if (ble_tj_conn_state.p_StateCB)
	        	{
		        	ble_tj_conn_state.p_StateCB(BLE_STATE_CONNECTED);

#if TEST_BLE_ACK
					BLE_callback_rssi_threshold_crossed(0, rssiCB);

					app_timer_start(m_test_timer_id, (APP_TIMER_CLOCK_FREQ * 2), NULL);
#endif
		        }
	  		}
	  	}
	}
}



/**@brief Function for initializing services that will be used by the application.
*/
static void services_init(void)
{
  uint32_t       err_code;
  ble_tj_init_t tj_init;
  
  memset(&tj_init, 0, sizeof(tj_init));
  
  tj_init.data_handler = tj_data_handler;
  
  err_code = ble_tj_init(&m_tj, &tj_init);
  APP_ERROR_CHECK(err_code);
}


/**@brief Function for handling the Security Request timer timeout.
*
* @details This function will be called each time the Security Request timer expires.
*
* @param[in] p_context  Pointer used for passing some arbitrary information (context) from the
*                       app_start_timer() call to the timeout handler.
*/
static void auth_timeout_handler(void * p_context)
{
  ret_code_t err_code;
  
  if (ble_tj_conn_state.conn_handle != BLE_CONN_HANDLE_INVALID)
  {
    // Initiate disconnection.
    NRF_LOG_DEBUG("Authentication failed ---> disconnect");
    err_code = sd_ble_gap_disconnect( ble_tj_conn_state.conn_handle, BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
    if (err_code != NRF_ERROR_INVALID_STATE)
    {
      APP_ERROR_CHECK(err_code);
    }
  }
}

#if TEST_BLE_ACK 
volatile int8_t rssi;

static void rssiCB(int8_t rssiv)
{
	NRF_LOG_DEBUG("RSSI threshold crossed %d", rssiv);
}

static void test_timeout_handler(void * p_context)
{
  // Send some buffer back to TJ server - this simulates command response
  //uint32_t    actual_length;
  //uint32_t       err_code;
  
  if (ble_tj_conn_state.connection_authenticated == true)
  {
    if (ble_tj_conn_state.conn_handle != BLE_CONN_HANDLE_INVALID)
    {
    	//BLE_get_rssi((int8_t *)&rssi);
#if 0
      actual_length = received_data_length;
      ble_tj_rsp_send(&m_tj, received_data, &received_data_length);
      APP_ERROR_CHECK(err_code);
      
      if (received_data_length != actual_length)
      {
        while(1)
          ;
      }
#endif
    }
  }
}
#endif

// The TX power shall be one of those:
// tx_power Radio transmit power in dBm (accepted values are -40,  -20, -16, -12, -8, -4, 0, and 4 dBm).
void BLE_do_init(uint16_t advertise_interval,
                 uint16_t  connection_interval_min,
                 uint16_t  connection_interval_max,
                 uint16_t  slave_latency,
                 uint16_t  supervision_timeout,
                 int16_t  tx_power)
{  
  ret_code_t err_code;
  
  // Create Security Request timer.
  err_code = app_timer_create(&m_auth_timer_id,
                              APP_TIMER_MODE_SINGLE_SHOT,
                              auth_timeout_handler);
  APP_ERROR_CHECK(err_code);
  
#if TEST_BLE_ACK 
  // Create Security Request timer.
  err_code = app_timer_create(&m_test_timer_id,
                              APP_TIMER_MODE_SINGLE_SHOT,
                              test_timeout_handler);
  
  APP_ERROR_CHECK(err_code);
#endif
  
  ble_tj_conn_state.conn_handle = BLE_CONN_HANDLE_INVALID;
  ble_tj_conn_state.ble_tj_max_data_len = BLE_GATT_ATT_MTU_DEFAULT - 3;
  ble_tj_conn_state.connection_authenticated = false;
  ble_tj_conn_state.alert_registered = false;
  ble_tj_conn_state.response_registered = false;
  ble_tj_conn_state.p_dataCB = 0;
  ble_tj_conn_state.p_StateCB = 0;
  ble_tj_conn_state.p_TxCompletedCB = 0;
  memset(&ble_tj_conn_state.peer_addr, 0, sizeof(ble_tj_conn_state.peer_addr));
  
  ble_stack_init();
  gap_params_init(connection_interval_min, connection_interval_max, slave_latency, supervision_timeout,  tx_power);
  gatt_init();
  services_init();
  advertising_init(advertise_interval, true);
  conn_params_init();
}


// The channel map includes 3 LSB's: each bit represents if the channel enabled (1) or disabled(0)
// The TX power shall be one of those:
// tx_power Radio transmit power in dBm (accepted values are -40,  -20, -16, -12, -8, -4, 0, and 4 dBm).

void BLE_do_adv_start(uint16_t channel_map, int16_t tx_power)
{
  uint32_t err_code;
  
  // if no channel enabled - nothing to ADV
  if (!channel_map)
  {
    return;
  }
  ble_tj_adv_state.ble_adv_map = channel_map;
  ble_tj_adv_state.ble_adv_tx_power = tx_power;
  
  err_code = sd_ble_gap_tx_power_set (ble_tj_adv_state.ble_adv_tx_power);
  APP_ERROR_CHECK(err_code);
  
  err_code = ble_advertising_start_with_map(&m_advertising, ble_tj_adv_state.ble_adv_map, ble_tj_adv_state.ble_adv_state);
  APP_ERROR_CHECK(err_code);
  
  ble_tj_adv_state.adv_active = true;
}


void BLE_register_callbacks(app_tj_data_handler_t dataCB, app_tj_state_handler_t stateCB, app_tj_tx_completed_handler_t txCompleteCB, app_tj_auth_handler_t authCB)
{
  ble_tj_conn_state.p_dataCB = dataCB;
  ble_tj_conn_state.p_StateCB = stateCB;
  ble_tj_conn_state.p_TxCompletedCB = txCompleteCB;
  ble_tj_conn_state.p_authHandlerCB = authCB;
}

void BLE_callback_rssi_threshold_crossed(int8_t threshold, app_tj_rssi_crossed_t cb)
{
	uint32_t err_code;
	
	if (cb == NULL)
	{
		// unregister from RSSI notifications if they are enabled
		if ((ble_tj_conn_state.p_rssiChangedCB) && (ble_tj_conn_state.conn_handle != BLE_CONN_HANDLE_INVALID))
		{
			sd_ble_gap_rssi_stop(ble_tj_conn_state.conn_handle);
			ble_tj_conn_state.p_rssiChangedCB = NULL;
		}
	}
	else
	{
		// Register in case we are connected
		if (ble_tj_conn_state.conn_handle != BLE_CONN_HANDLE_INVALID)
		{
			ble_tj_conn_state.p_rssiChangedCB = cb;
			err_code = sd_ble_gap_rssi_start(ble_tj_conn_state.conn_handle, threshold, 3);

			if (err_code == NRF_ERROR_INVALID_STATE)
			{
				ble_tj_conn_state.p_rssiChangedCB = 0;
			}
			else if (err_code == BLE_ERROR_INVALID_CONN_HANDLE)
			{
				APP_ERROR_CHECK(err_code);
			}
		}
	}
}


