/******************************************************************************* 

  motor.c 

  This file contains the source code for the motor class. 

  omri-todo:
  - select rotation direction 
  - check power with eli 
  - v - verify adc works - sampling rate ok
  - v - erify adc works - acq time works ok 
  - v - verify adc works - splitting to 3 channnels works (2/7 had one occurance of skip channel)
  - verify adc works - sampling every 8 pulses and calculating fast enough to stop on time in doser knee 
  - v - verify adc works - switch to 3us acq and 5us sampling rate and verify operability 
  - ensure low power in between trains with eli 
  - port the entire motor logic 
  - port the entire doser logic 
  - v - buzzer not working 
  - verify trains from 7 to 900 pulses works fine 
  - update number of pulses from the sw irq as the train progresses to avoid over delivery 
  - check why PRE_DEDUCTION 0 not working 

*******************************************************************************/

#include "nrf_delay.h"
#include "nrf_gpio.h"
#include "nrf_drv_saadc.h"

#define ADC_TEST        1 // 1 or 2 

static volatile uint32_t train_dummy;
#define NRF_SAADC_EVENTS_CLEAR(SAADC_REG_NAME) NRF_SAADC->SAADC_REG_NAME = 0x0UL; train_dummy = NRF_SAADC->SAADC_REG_NAME; (void)train_dummy;

static int adc_buffer_count;
static int adc_buffer_count_target;

static nrf_saadc_value_t adc_buffer[2][333];

__root uint32_t timing_debug_buff[256];
__root uint8_t timing_debug_i = 0;

void timing_debug(uint32_t val)
{
  static uint32_t p = 0;
  NRF_TIMER2->TASKS_CAPTURE[0] = 1;
  uint32_t now = NRF_TIMER2->CC[0];  
  if ( !p )
  {
    p = now;
  }
  uint32_t delta = now - p;
  p = now;
  timing_debug_buff[timing_debug_i++] = (delta & 0xFFFF) | (val<<16);
}


/******************************************************************************* 

  SAADC_IRQHandler

*******************************************************************************/
void SAADC_IRQHandler(void)
{
  if (nrf_saadc_event_check(NRF_SAADC_EVENT_STARTED))
  {
    adc_buffer_count++;

    timing_debug(adc_buffer_count);
    
    // clear event
    NRF_SAADC_EVENTS_CLEAR(EVENTS_STARTED)
      
    // set its pointer to the desired location before starting again 
    int delta = NRF_SAADC->RESULT.MAXCNT*sizeof(uint16_t);
    NRF_SAADC->RESULT.PTR += delta;
    
    // once the last buffer is trigerred - hook EVENTS_END to TASKS_STOP instead of to TASKS_START 
    if ( adc_buffer_count==adc_buffer_count_target )
    {
      NRF_PPI->CHENCLR = PPI_CHENCLR_CH1_Clear << PPI_CHENCLR_CH1_Pos;
      NRF_PPI->CH[1].EEP = (uint32_t) &NRF_SAADC->EVENTS_END;
      NRF_PPI->CH[1].TEP = (uint32_t) &NRF_SAADC->TASKS_STOP;
      NRF_PPI->CHENSET = PPI_CHENSET_CH1_Enabled << PPI_CHENSET_CH1_Pos;
    }
    
    if ( adc_buffer_count>1 )
    {
      // todo - do processing of adc data for this pulse (buffer)
    }
  }

  if (nrf_saadc_event_check(NRF_SAADC_EVENT_STOPPED))
  {
    // last pulse 
    adc_buffer_count++;
    timing_debug(adc_buffer_count);
    
    // clear event and disable interrupts 
    NRF_SAADC_EVENTS_CLEAR(EVENTS_STOPPED)
    NRF_SAADC->INTENCLR = NRF_SAADC_INT_ALL; 
      
    // disable the adc 
    NRF_SAADC->ENABLE = (SAADC_ENABLE_ENABLE_Disabled << SAADC_ENABLE_ENABLE_Pos);
    NRF_SAADC->CH[0].PSELP = NRF_SAADC_INPUT_DISABLED;
    NRF_SAADC->CH[1].PSELP = NRF_SAADC_INPUT_DISABLED;
    NRF_SAADC->CH[2].PSELP = NRF_SAADC_INPUT_DISABLED;
    
    // disaple the ppi   
    NRF_PPI->CHENCLR = (PPI_CHENCLR_CH0_Clear << PPI_CHENCLR_CH0_Pos) | (PPI_CHENCLR_CH1_Clear << PPI_CHENCLR_CH1_Pos);
    
    // disable the timers   
    NRF_TIMER1->TASKS_STOP = 1;
    NRF_TIMER1->SHORTS = 0;
    NRF_TIMER1->TASKS_SHUTDOWN = 1;
    
    NVIC_ClearPendingIRQ(SAADC_IRQn);
    NVIC_DisableIRQ(SAADC_IRQn);
    
    // todo - do processing of adc data for the last pulse (buffer)
  }
}

void a2d_test(void)
{
  adc_buffer_count = 0;
  adc_buffer_count_target = 10;

  // free runningtimer for debug 
  NRF_TIMER2->BITMODE = TIMER_BITMODE_BITMODE_32Bit;
  NRF_TIMER2->MODE = (TIMER_MODE_MODE_Timer << TIMER_MODE_MODE_Pos); 
  NRF_TIMER2->PRESCALER = 0; 
  
  // timer that clocks the adc once every 18us 
  NRF_TIMER1->MODE = (TIMER_MODE_MODE_Timer << TIMER_MODE_MODE_Pos); 
  NRF_TIMER1->PRESCALER = 4; // 1mhz --> 1us 
  NRF_TIMER1->CC[0] = 18; 
  NRF_TIMER1->SHORTS = TIMER_SHORTS_COMPARE0_CLEAR_Msk;

  // connect the timer to adc sampling 
  NRF_PPI->CH[0].EEP = (uint32_t) &NRF_TIMER1->EVENTS_COMPARE[0];
  NRF_PPI->CH[0].TEP = (uint32_t) &NRF_SAADC->TASKS_SAMPLE;

  // connect end to start 
  NRF_PPI->CH[1].EEP = (uint32_t) &NRF_SAADC->EVENTS_END;
  NRF_PPI->CH[1].TEP = (uint32_t) &NRF_SAADC->TASKS_START;

  // enable ppi 
  NRF_PPI->CHENSET = (PPI_CHENSET_CH0_Enabled << PPI_CHENSET_CH0_Pos) | 
                     (PPI_CHENSET_CH1_Enabled << PPI_CHENSET_CH1_Pos);

  // configure the adc to sample 3 channels 111 samples per channel (total 333 samples)
  NRF_SAADC->ENABLE = (SAADC_ENABLE_ENABLE_Enabled << SAADC_ENABLE_ENABLE_Pos);
  NRF_SAADC->RESOLUTION = NRF_SAADC_RESOLUTION_12BIT;
  NRF_SAADC->OVERSAMPLE = NRF_SAADC_OVERSAMPLE_DISABLED;
  NRF_SAADC_EVENTS_CLEAR(EVENTS_END)
  NRF_SAADC_EVENTS_CLEAR(EVENTS_STARTED)
  for ( int i=0; i<3; i++ )
  {
    NRF_SAADC->CH[i].PSELN = NRF_SAADC_INPUT_DISABLED;
    NRF_SAADC->CH[i].CONFIG =
            ((NRF_SAADC_RESISTOR_DISABLED   << SAADC_CH_CONFIG_RESP_Pos)   & SAADC_CH_CONFIG_RESP_Msk)
            | ((NRF_SAADC_RESISTOR_DISABLED << SAADC_CH_CONFIG_RESN_Pos)   & SAADC_CH_CONFIG_RESN_Msk)
            | ((NRF_SAADC_GAIN1_4           << SAADC_CH_CONFIG_GAIN_Pos)   & SAADC_CH_CONFIG_GAIN_Msk)
            | ((NRF_SAADC_REFERENCE_VDD4    << SAADC_CH_CONFIG_REFSEL_Pos) & SAADC_CH_CONFIG_REFSEL_Msk)
            | ((NRF_SAADC_ACQTIME_3US       << SAADC_CH_CONFIG_TACQ_Pos)   & SAADC_CH_CONFIG_TACQ_Msk)
            | ((NRF_SAADC_MODE_SINGLE_ENDED << SAADC_CH_CONFIG_MODE_Pos)   & SAADC_CH_CONFIG_MODE_Msk)
            | ((NRF_SAADC_BURST_DISABLED    << SAADC_CH_CONFIG_BURST_Pos)  & SAADC_CH_CONFIG_BURST_Msk);
  }
  NRF_SAADC->CH[0].PSELP = NRF_SAADC_INPUT_AIN1;
  NRF_SAADC->CH[1].PSELP = NRF_SAADC_INPUT_AIN2;
  NRF_SAADC->CH[2].PSELP = NRF_SAADC_INPUT_AIN3;
  NRF_SAADC->RESULT.PTR = (uint32_t)adc_buffer[0];
  NRF_SAADC->RESULT.MAXCNT = 9; // 333 ;
  NRF_SAADC_EVENTS_CLEAR(EVENTS_STARTED)
  NRF_SAADC_EVENTS_CLEAR(EVENTS_STOPPED)
  NRF_SAADC_EVENTS_CLEAR(EVENTS_END)
    
#if ADC_TEST==2
  NRF_SAADC->INTENCLR = NRF_SAADC_INT_ALL;
  NVIC_SetPriority(SAADC_IRQn, SAADC_CONFIG_IRQ_PRIORITY);
  NVIC_ClearPendingIRQ(SAADC_IRQn);
  NVIC_EnableIRQ(SAADC_IRQn);
  NRF_SAADC->INTENSET = NRF_SAADC_INT_STARTED;
  NRF_SAADC->INTENSET = NRF_SAADC_INT_STOPPED;
#endif
    
  // trigger the timer that clocks the adc and the adc 
  NRF_SAADC->TASKS_START = 1;
  NRF_TIMER1->TASKS_START = 1;  
  NRF_TIMER2->TASKS_START = 1;  

  timing_debug(0);
  
#if ADC_TEST==1
  // wait till adc has started and place the next buffer so it will be ready at EVENTS_END/TASKS_START
  while (NRF_SAADC->EVENTS_STARTED==0);
  int delta = NRF_SAADC->RESULT.MAXCNT*sizeof(uint16_t);
  NRF_SAADC->RESULT.PTR += delta;
  
  // loop waiting for each buffer to end and then process by SW 
  while ( adc_buffer_count < adc_buffer_count_target )
  {
    // wait until adc ends its sampling and automatically starts the next buffer 
    NRF_SAADC_EVENTS_CLEAR(EVENTS_STARTED)
    while (NRF_SAADC->EVENTS_END==0);
    NRF_SAADC_EVENTS_CLEAR(EVENTS_END)
    while (NRF_SAADC->EVENTS_STARTED==0);

    // set its pointer to the desired location before starting again 
    int delta = NRF_SAADC->RESULT.MAXCNT*sizeof(uint16_t);
    NRF_SAADC->RESULT.PTR += delta;
    adc_buffer_count++;
    timing_debug(adc_buffer_count);

    // todo - here we have time to porcess the data in a double buffer mode 
    
    // once the last buffer is trigerred - hook EVENTS_END to TASKS_STOP instead of to TASKS_START 
    if ( adc_buffer_count==adc_buffer_count_target-1 )
    {
      NRF_PPI->CHENCLR = PPI_CHENCLR_CH1_Clear << PPI_CHENCLR_CH1_Pos;
      NRF_PPI->CH[1].EEP = (uint32_t) &NRF_SAADC->EVENTS_END;
      NRF_PPI->CH[1].TEP = (uint32_t) &NRF_SAADC->TASKS_STOP;
      NRF_PPI->CHENSET = PPI_CHENSET_CH1_Enabled << PPI_CHENSET_CH1_Pos;
      while ( NRF_SAADC->EVENTS_STOPPED==0 );
      adc_buffer_count++;
    }
  }

  // disable the adc 
  timing_debug(adc_buffer_count);
  NRF_SAADC->ENABLE = (SAADC_ENABLE_ENABLE_Disabled << SAADC_ENABLE_ENABLE_Pos);
  NRF_SAADC->CH[0].PSELP = NRF_SAADC_INPUT_DISABLED;
  NRF_SAADC->CH[1].PSELP = NRF_SAADC_INPUT_DISABLED;
  NRF_SAADC->CH[2].PSELP = NRF_SAADC_INPUT_DISABLED;

  // disaple the ppi   
  NRF_PPI->CHENCLR = (PPI_CHENCLR_CH0_Clear << PPI_CHENCLR_CH0_Pos) | 
                     (PPI_CHENCLR_CH1_Clear << PPI_CHENCLR_CH1_Pos);

  // disable the timers   
  NRF_TIMER1->TASKS_STOP = 1;
  NRF_TIMER1->SHORTS = 0;
  NRF_TIMER1->TASKS_SHUTDOWN = 1;

  NRF_TIMER2->TASKS_STOP = 1;
  NRF_TIMER2->SHORTS = 0;
  NRF_TIMER2->TASKS_SHUTDOWN = 1;
#endif  
}
